import json
import websocket

channels_ws = []


class Feed(object):
    def __init__(self):
        self.url = 'ws://192.168.3.9:9990'
        self.ws = None

    def on_open(self, ws):
        print('A new WebSocketApp is opened!')
        sub_param = {"op": "subscribe", "args": channels_ws}
        sub_str = json.dumps(sub_param)
        ws.send(sub_str)
        print("Following Channels are subscribed!")
        print(channels_ws)

    def on_data(self, ws, string, type, continue_flag):
        """

        Args:
            ws: this class object
            string: a utf8 string which we get from the server
            type: data type, ABNF.OPCODE_TEXT or ABNF.OPCODE_BINARY
            continue_flag:

        Returns:

        """
        ...

    def on_message(self, ws, message):
        """
        当接收到消息时的回调函数
        Args:
            ws: the WebSocketApp object
            message: 从服务器接收到的消息

        Returns:

        """
        result = message
        print(result)

    def on_error(self, ws, error):
        """
        当遇到错误时的回调函数
        Args:
            ws:
            error: exception object

        Returns:

        """
        print(error)

    def on_close(self, ws, close_status_code, close_msg):
        """
        连接关闭时的回调函数

        Args:
            ws:
            close_status_code:
            close_msg:

        Returns:

        """
        print("连接已经关闭")

    def start(self):
        self.ws = websocket.WebSocketApp(
            self.url,
            on_open=self.on_open,
            on_message=self.on_message,
            on_data=self.on_data,
            on_error=self.on_error,
            on_close=self.on_close,

        )
        self.ws.run_forever()


if __name__ == "__main__":
    feed = Feed()
    feed.start()



import datetime
import traceback
from typing import Any
import importlib.metadata
from typing import Union, Type
import pandas as pd
from sqlalchemy import create_engine, MetaData, Column, inspect, Table, String, text, insert, delete, update, select, \
    func, asc, desc, BOOLEAN
from sqlalchemy.dialects.mysql import INTEGER, DOUBLE, BIGINT, VARCHAR, CHAR, TEXT, DATETIME, DATE
from sqlalchemy.engine import Engine, Row
from sqlalchemy.exc import PendingRollbackError
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from sqlalchemy.sql import Select

from yangke.common.config import logger


def get_sqlalchemy_version() -> str:
    """
    获取 SQLAlchemy 的版本号。
    """
    try:
        # Python 3.8+ 推荐使用 importlib.metadata
        version = importlib.metadata.version("sqlalchemy")
    except ImportError:
        # 兼容旧版本 Python
        import pkg_resources
        version = pkg_resources.get_distribution("sqlalchemy").version
    return version


def import_by_version(module_path: str, class_name: str) -> Union[Type, None]:
    """
    根据模块路径和类名动态导入类。
    """
    module = __import__(module_path, fromlist=[class_name])
    return getattr(module, class_name)


def import_sqlalchemy_components() -> dict:
    """
    根据 SQLAlchemy 版本动态导入所需的组件。
    """
    version = get_sqlalchemy_version()
    major_version = int(version.split(".")[0])  # 获取主版本号

    components1 = {}

    if major_version >= 2:
        # SQLAlchemy 2.0+ 的导入路径
        components1["Engine"] = import_by_version("sqlalchemy", "Engine")
        components1["QueuePool"] = import_by_version("sqlalchemy", "QueuePool")
        components1["Row"] = import_by_version("sqlalchemy.engine", "Row")
        components1["Double"] = import_by_version("sqlalchemy", "Double")
    elif major_version == 1:
        # SQLAlchemy 1.4.x 的导入路径
        components1["Engine"] = import_by_version(
            "sqlalchemy.engine", "Engine")
        components1["QueuePool"] = import_by_version(
            "sqlalchemy.pool", "QueuePool")
        components1["Row"] = import_by_version(
            "sqlalchemy.engine.result", "Row")
        components1["Double"] = import_by_version(
            "sqlalchemy", "Float")  # 1.4.x 中使用 Float 替代 Double
    else:
        raise ImportError(f"Unsupported SQLAlchemy version: {version}")

    return components1


# 动态导入 SQLAlchemy 组件
components = import_sqlalchemy_components()
# Engine = components["Engine"]
QueuePool = components["QueuePool"]
# Row = components["Row"]
Double = components["Double"]

# # 创建基类, 使用ORM方式操作数据库是继承该类
Base = declarative_base()
# noinspection all
修改时间记录表 = 'modify_time'


class ModifyTime(Base):
    """
    用来记录数据库表格最后更新时间的表格，因为目前MySql和Sqlite数据库均无法查询表格的最后更新时间，因此使用数据库表格记录所有表格的最后更新
    时间
    """
    __tablename__ = 修改时间记录表
    table = Column(String(50, collation='utf8mb4_bin'),
                   nullable=False, primary_key=True)  # 设置collation为区分大小写
    datetime = Column(DATETIME, nullable=False)

    def __init__(self, table, date_time):
        super(ModifyTime, self).__init__()
        self.table = table  # 表名
        self.datetime = date_time  # 表的最后修改时间

    def __repr__(self):
        return f"ModifyTime(table={self.table}, 修改时间={self.datetime})"


class YkTable(Table):
    def __init__(self, name, metadata, columns=None):
        if columns is None:
            columns = []
        super().__init__(name, metadata, *columns)


class YkColumn(Column):
    inherit_cache = True

    def __init__(self, name, dtype, primary_key=False, nullable=False):
        """
        Sqlite数据库不支持foreignKey关键字，YkColumn的内容默认区分大小写
        :param name: 列名
        :param dtype: 数据类型
        :nullable: 是否可以为空
        """
        if (isinstance(dtype, String) or isinstance(dtype, VARCHAR) or
                isinstance(dtype, CHAR) or isinstance(dtype, TEXT)):
            dtype.collation = 'utf8mb4_bin'
        super().__init__(name, dtype, primary_key=primary_key, nullable=nullable)


class SqlOperator:
    def __init__(self, engine: Engine = None):
        """
        SQL数据库操作类，同时支持MySql和Sqlite数据库，使用示例：
        engine = create_engine("mysql+mysqlconnector://root:password@localhost:3306", pool_recycle=7200)
        engine = create_engine('sqlite:///stocks.db', echo=True)
        create_engine(f'mysql+pymysql://{mysql_user}:{mysql_password}@{mysql_host}:{mysql_port}/{mysql_db}',
                      poolclass=QueuePool,
                      pool_size=5,
                      pool_timeout=30,  # 连接池中没有可用连接时的等待时间，超过该时间还没有可用连接，则报错
                      pool_recycle=3600,  # 连接池中的连接1小时后会回收
                      pool_pre_ping=True,  # 每次连接请求时，先测试返回的连接是否还存活，不存活则从连接池中取下一个连接
                      )  # 使用连接池
        so = SqlOperator(engine=engine)
        然后可以调用so的增删改查等方法，所有方法均同时支持MySql和Sqlite数据库

        可能遇到的问题：
        1.sqlalchemy.exc.InterfaceError: (mysql.connector.errors.InterfaceError) 2003: Can't connect to MySQL server on
        'disk.yangke.site:3306' (10060 由于连接方在一段时间后没有正确答复或连接的主机没有反应，连接尝试失败。)
        解决方法： 不使用mysqlconnector连接，pip install mysql后使用：create_engine("mysql://root:password@localhost:3306")
        """
        # 创建数据库引擎，ps:这里并没有连接具体数据库
        # engine = create_engine("mysql+mysqlconnector://root:password@localhost:3306/db?charset=utf8", pool_recycle=7200)

        # 连接时，如果报Character set 'utf8' unsupported错误，且数据库文件中没有非ascii字符，可以尝试切换charset的值。
        # 已知charset可取值包括(ascii, utf8, utf8mb4, gbk, cp1250)，可以执行show collation语句查看mysql数据库支持的字符集。

        self.engine = engine
        # self.insp = inspect(self.engine)  # 在需要时inspect，而不是开始就inspect
        self.meta_data = MetaData()  # 兼容sqlalchemy 2.0版本的写法，bind参数写在reflect()方法中
        self.meta_data.reflect(bind=engine)  # 需要的时候在reflect，否则当数据库中表多的话，耗时很长
        self.SessionMaker = sessionmaker(
            bind=engine)  # 使用sessionmaker保持数据库会话连接
        self.base = Base
        # self.connect = self.engine.connect() // Connection随用随关，不能长时间保持连接
        self.create_all_base_table()  # 创建修改时间记录表
        self._table_cache = {}  # 新增表名缓存
        self._last_refresh = None  # 最后刷新时间

    def create_table(self, table_name, columns):
        """
        创建数据库表，如果同名表存在，会报错，SqlAlchemy 2.0.36已测试
        示例1：
        self.create_table(table_name=f"daily{symbol}",
                          columns=[
                              YkColumn('trade_date', DATE(), primary_key=True),
                              YkColumn('open', Float()),
                              YkColumn('high', Float()),
                              YkColumn('low', Float()),
                              YkColumn('close', Float()),
                              YkColumn('vol', Float()),
                              YkColumn('amount', Float()),
                          ])
        """
        table = Table(table_name, self.meta_data, *columns)
        logger.debug(f"创建表格{table_name}")
        with self.engine.connect() as connection:
            table.create(bind=connection)  # bind=self.engine时，部分情况下会卡住

    def create_all_base_table(self):
        """
        创建所有继承自Base类的Python类的映射数据库表。也就是说，只要定义了继承自本模块中Base类的类，则该类会被映射成一个数据库表，
        本方法会自动创建所有映射的数据库表。本方法主要用于创建修改时间记录表，如果创建用户的专有表，建议使用self.create_table()方法
        """
        self.base.metadata.create_all(self.engine)

    def get_type_of_column(self, table_name=None, column_name=None):
        """
        获取mysql表中字段的类型，如果不设置column_name则返回所有的字段类型
        :param table_name:
        :param column_name: 为空则依次返回所有列的类型，封装为一个列表
        :return:
        """
        cols = inspect(self.engine).get_columns(table_name)
        res = None
        if column_name is None:
            res = {}
            for col in cols:
                res.update({col["name"]: col["type"]})
        else:
            for col in cols:
                if col["name"] == column_name:
                    res = col["type"]
                    break
        return res

    def get_column_names(self, table_name):
        """
        获取表格中的列名，返回列名列表
        """
        cols = inspect(self.engine).get_columns(table_name)
        res = [col["name"] for col in cols]
        return res

    def get_update_time_of_table(self, table_name):
        """
        获取表的最后更新时间
        """
        session = self.SessionMaker()
        try:
            if self.has_table(table_name):
                # 查询modifyTime表，modifyTime表中记录了所有表格的最后更新时间
                if self.has_table(修改时间记录表):
                    table1: Table = self.get_table(修改时间记录表)
                    select1: Select = table1.select().where(
                        table1.c.table == table_name)  # 构建一个选择语句
                    res = session.execute(select1)  # 执行选择语句并获取结果
                    res = res.fetchone()
                    # res = session.query(table1).filter(table1.table == table_name).first()

                    if res is None:
                        return None
                    return res.datetime  # 该写法兼容sqlalchemy1.4和2.0
                    # return res["datetime"]  # 该写法不兼容sqlalchemy2.0
                else:
                    return None
            else:
                return None
        except PendingRollbackError:
            try:
                session.rollback()
            except Exception:
                logger.warning("回滚失败")
        finally:
            session.close()

    def update_update_time_of_table(self, table_name):
        """
        更新表的最后更新时间
        """
        session = self.SessionMaker()
        try:
            if not self.has_table(修改时间记录表):
                self.create_all_base_table()

            now = datetime.datetime.now()
            if self.exists_in_table(table_name=修改时间记录表, col_name='table', value=table_name):
                self.update_item(table_name=修改时间记录表, conditions={
                                 "table": table_name}, values={"datetime": now})
            else:
                self.insert_item(table_name=修改时间记录表, values=[
                                 table_name, now], col_names=['table', 'datetime'])
                # session.add(ModifyTime(table_name, now))
            session.commit()
        except PendingRollbackError:
            try:
                session.rollback()  # 尝试回滚，但可能已经有一个回滚在进行中
            except Exception as rollback_exc:
                # 记录回滚失败的信息
                print(f"回滚失败: {rollback_exc}")
        except Exception as e:
            # 处理其他数据库异常
            session.rollback()
            print(f"发生数据库错误: {e}")
        finally:
            session.close()

    def exists_in_table(self, table_name: str = None, col_name: str = None, value: str = None,
                        condition_dict: dict = None,
                        return_result: bool = False):
        """
        表tableName中是否存在列col_name的值位value的行

        :param table_name: 表名
        :param col_name: 列名
        :param value: 列的值
        :param condition_dict: 查询的键值对字典值，优先于col_name和value传入的值，即会覆盖col_name和value传入的值
        :param return_result: 是否需要返回查找到的数据行，如果为真，则返回所有符合查找条件的数据行
        :return:
        """
        if return_result:
            first_or_all = 'all'
        else:
            first_or_all = 'first'

        condition_dict = condition_dict or {}
        if col_name is not None and value is not None:
            condition_dict.update({col_name: value})

        res = self.select_item(table_name, condition_dict,
                               first_or_all=first_or_all)
        if res is not None:
            return True
        else:
            return False

    def select_in_table(self, table_name=None, condition_dict: dict = None, result_col: list | str = None, limit=10,
                        offset=0,
                        fuzzy=False, first_or_all="first", result_type=None, cls=None, **kwargs) -> Row | list | Any:
        """
        查

        精确查询，设置fuzzy为True，且condition_dict中的value为字符串值
        模糊查询，设置fuzzy为False，日期列不能使用模糊查询，只能使用范围查询
        范围查询，设置fuzzy为True，且condition_dict中的value为长度为2的列表，列表第一、二项分别为范围下、上限，且包含上下限

        当result_type=="json"时，返回的是一个list(dict)的json对象，即[{col1: value1, col2: value2,...}, ...}的json对象
        列表的每一项对应一条匹配的查询结果
        每一项的字典分别是{列名：值}

        kwargs={"date_format": "%Y-%m-%d %H:%M:%S"} 如果mysql中存在日期列，需要将日期转换为字符串，该参数定义日期字符串格式

        示例1：
        fetch = self.sql.select_in_table(cls=Holiday, condition_dict={"calendarDate": day_datetime},
                                         result_col=['isOpen'])

        :param table_name: 当使用传统查询方式时，需要传入数据库表名
        :param cls: 当使用ORM模型时，只需要传入数据库表在python中对应的映射类名，如果通过该方法查询，则查询到的数据会被自动转换为cls对象
        :param condition_dict:
        :param result_col: 不传入或传入空列表，则返回数据库中所有列
        :param limit:
        :param offset:
        :param fuzzy: 是否模糊查询
        :param first_or_all: 返回满足条件的第一个还是所有，支持"first", "all",
        :param result_type: 返回类型，如果为json，则返回为json格式的字符串
        :return: None或查询的列值列表或数据条的列表或sqlalchemy.engine.Row对象，出错时返回None，如列明不存在等；否则返回一个tuple类型的数据，长度为0表示未查询到满足条件的数据
        """
        # ---------------------- 如果使用table_name查询，则构建Table对象 ------------------------
        # Table对象可以当cls一样使用
        if cls is None:
            if self.has_table(table_name):
                table: Table = self.get_table(table_name)
                if table is None:
                    return None
                cls = table
        # ---------------------- 如果使用table_name查询，则构建Table对象 ------------------------
        session = self.SessionMaker()
        try:
            # ----------------------- 根据条件字典，构建查询条件 --------------------------------
            if condition_dict is None:  # 查询表格中某列的所有值
                if result_col is None:
                    result_col = self.get_column_names(table_name)
                if isinstance(result_col, str):
                    res = session.execute(
                        select(cls.c.get(result_col))).fetchall()
                    res = [i._data[0] for i in res]
                else:  # list
                    n = len(result_col)
                    _1 = []
                    for i in range(n):
                        _1.append(f"cls.c.get(result_col[{i}])")
                    _2 = ",".join(_1)

                    _3 = f"select({_2})"
                    try:
                        _4 = eval(_3)
                        res = session.execute(_4).fetchall()
                        res = [i._data for i in res]
                    except:
                        traceback.print_exc()
                        logger.error(f"检查返回列的名称是否正确：{result_col=}")
                        res = None
                if result_type == "json":
                    res = [
                        dict(zip([str(column.name) for column in table.c], list(item))) for item in res]
                return res

            _1 = []
            vals = []
            i = 0
            for k, v in condition_dict.items():
                vals.append(v)  # 为了兼容v为datetime或date类型的参数
                _1.append(f"cls.c.{k}==vals[{i}]")
                i += 1
            _2 = ",".join(_1)

            _3 = f"select(cls).where({_2})"
            items = session.execute(eval(_3)).fetchall()
            if len(items) == 0 or items is None:
                return None
            if first_or_all == "first":
                item = items[0]
            else:
                item = items

            # ------------------------ 如果指定了返回的数据列，则取出数据列并返回 -----------------------
            if isinstance(item, str):
                item = [item]
            if item is None or len(item) == 0:
                return item
            else:
                try:
                    # item是个sqlalchemy.engine.row.Row对象，如读取Holiday表时会出现该情况
                    if isinstance(item, Row):
                        if result_col is None:
                            if result_type == 'json':
                                return dict(zip([str(column.name) for column in table.c], list(item)))
                            else:
                                return item
                        elif isinstance(result_col, str):
                            # sqlalchemy2.0中，item.__getattr__(result_col)语法是正确的
                            # sqlalchemy1.4中，item.__getitem__(result_col)语法是正确的
                            # 该写法兼容sqlalchemy2.0
                            res = eval(f"item.{result_col}")
                            return res
                        elif isinstance(result_col, list):
                            if len(result_col) == 1:
                                # 该写法兼容sqlalchemy2.0
                                res = eval(f"item.{result_col[0]}")
                                return res
                            else:
                                return [getattr(item, col) for col in result_col]

                    logger.debug(f"{item=}, {type(item)=}")
                    # noinspection all
                    _ = [getattr(item, col)
                         for col in item]  # table_name和cls两种方法都适用
                    if len(_) == 1:
                        _ = _[0]
                    return _
                except AttributeError:
                    logger.error(item)
            # ------------------------ 如果指定了返回的数据列，则取出数据列并返回 -----------------------
        except PendingRollbackError:
            try:
                session.rollback()  # 尝试回滚，但可能已经有一个回滚在进行中
            except Exception as rollback_exc:
                # 记录回滚失败的信息
                print(f"回滚失败: {rollback_exc}")
        finally:
            session.close()

    def select_item(self, table_name=None, condition_dict: dict = None, result_col: list | str = None, limit=10,
                    offset=0, first_or_all="first",
                    fuzzy=False, result_type=None):
        """
        查

        精确查询，设置fuzzy为True，且condition_dict中的value为字符串值
        模糊查询，设置fuzzy为False，日期列不能使用模糊查询，只能使用范围查询
        范围查询，设置fuzzy为True，且condition_dict中的value为长度为2的列表，列表第一、二项分别为范围下、上限，且包含上下限

        当result_type=="json"时，返回的是一个list(dict)的json对象，即[{col1: value1, col2: value2,...}, ...}的json对象
        列表的每一项对应一条匹配的查询结果
        每一项的字典分别是{列名：值}

        kwargs={"date_format": "%Y-%m-%d %H:%M:%S"} 如果mysql中存在日期列，需要将日期转换为字符串，该参数定义日期字符串格式

        示例1：
        fetch = self.sql.select_in_table(cls=Holiday, condition_dict={"calendarDate": day_datetime},
                                         result_col=['isOpen'])

        :param table_name: 当使用传统查询方式时，需要传入数据库表名
        :param condition_dict:
        :param result_col: 不传入或传入空列表，则返回数据库中所有列
        :param limit:
        :param offset:
        :param fuzzy: 是否模糊查询
        :param first_or_all: 返回满足条件的第一个还是所有，支持"first", "all",
        :param result_type: 返回类型，如果为json，则返回为json格式的字符串
        :return: None或查询的列值列表或数据条的列表或sqlalchemy.engine.Row对象，出错时返回None，如列明不存在等；否则返回一个tuple类型的数据，长度为0表示未查询到满足条件的数据
        """
        return self.select_in_table(table_name=table_name, condition_dict=condition_dict, result_col=result_col,
                                    limit=limit, offset=offset,
                                    fuzzy=fuzzy, first_or_all=first_or_all, result_type=result_type)

    def update_item(self, table_name, conditions: dict, values):
        """
        self.update_item('user', {"id": 1}, values={"name": "Tom"})

        兼容sqlalchemy 1.4及2.0以上版本。
        query.filter()或query.filter_by()的写法无法同时兼容sqlalchemy 1.4和2.0版本

        Parameters
        ----------
        table_name
        conditions: 数据表的索引列的值，必须唯一确定某一个记录
        values: 数据表中其他列的值，可以只设置部分列的值

        Returns
        -------

        """
        table: Table = self.get_table(table_name)
        update_obj = update(table)
        for col_title, col_value in conditions.items():
            update_obj = update_obj.where(
                table.c.get(col_title) == col_value)  # 兼容写法

        dict_str = "{"
        objs = []
        i = 0
        for title, value in values.items():
            # if isinstance(value, datetime):
            #     dict_str = f"{dict_str}table.c.{title}: value,"
            objs.append(value)
            # 这样的写法是为了解决value是datetime或date类型时的兼容写法
            dict_str = f"{dict_str} table.c.{title}: objs[{i}],"
            i += 1
        dict_str += "}"
        values = eval(dict_str)
        update_obj = update_obj.values(values)
        session = self.SessionMaker()
        try:
            session.execute(update_obj)
            session.commit()
        except PendingRollbackError:
            try:
                session.rollback()  # 尝试回滚，但可能已经有一个回滚在进行中
            except Exception as rollback_exc:
                # 记录回滚失败的信息
                print(f"回滚失败: {rollback_exc}")
        finally:
            session.close()

    def get_primary_col(self, table_name):
        """
        获取表的主键

        Parameters
        ----------
        table_name

        Returns
        -------

        """
        table = self.get_table(table_name)
        if isinstance(table, Table):
            # noinspection all
            for col in table.columns:
                if col.primary_key:
                    return col
        return None

    def insert_dataframe(self, table_name, df: pd.DataFrame, if_exists="append", index=False, dtype=None):
        """
        将DataFrame对象写入数据库中。兼容mysql和sqlite。df对象可以包含数据库中不存在的列，该方法会忽略这些列。同时该方法会记录表的最后更新时间

        pandas的to_sql方法存在以下问题：1.表格存在时，如果使用replace，则新创建的表的主键会丢失；2.append时会报重复键错误；

        批量插入数据时，
        简单说明：同样添加10W行数据，插入时间比
        1，Session.add(obj)   使用时间：6.89754080772秒
        2，Session.add(obj)注意：手动添加了主键id   使用时间：4.09481811523秒
        3，bulk_save_objects([obj1, obj2])   使用时间：1.65821218491秒
        4，bulk_insert_mappings(DBmodel, [dict1, dict2])  使用时间： 0.466513156781秒
        5，SQLAlchemy_core(DBmodel.__table__insert(), [dict1, dict2]) 使用时间：0.21024107933秒
        6，直接执行execute(str_sql_insert)  直接执行sql插入语句 使用时间：0.137335062027秒
        ————————————————
        版权声明：本文为CSDN博主「DHogan」的原创文章，遵循CC 4.0 BY-SA版权协议，转载请附上原文出处链接及本声明。
        原文链接：https://blog.csdn.net/dongyouyuan/article/details/79236673

        Parameters
        ----------
        table_name
        df
        if_exists: 参数参加pandas.DataFrame().to_sql()方法，可取值append, replace, fail
        index
        dtype : 指定df中各列的数值类型

        Returns
        -------

        """

        def slice_str(long_string):
            if len(long_string) > 200:
                short_string_with_ellipsis = long_string[:200] + "..."
            else:
                short_string_with_ellipsis = long_string
            return short_string_with_ellipsis

        session = self.SessionMaker()
        try:
            # 逐条插入，只要表存在，则不要使用pandas的to_sql方法，因为to_sql方法会覆盖原表，导致丢失主键
            if self.has_table(table_name):
                primary_col = self.get_primary_col(table_name)
                primary_key = primary_col.name
                items_exist = self.select_item(
                    table_name, result_col=primary_key)  # 获取数据库中表现有的数据记录的唯一标记
                primary_type = self.get_type_of_column(table_name, primary_key)
                if isinstance(items_exist, list):
                    table: Table = self.get_table(table_name)
                    # ---------------------- 判断需要插入的和需要更新的数据 ---------------------------
                    # 有些股票的交易量为nan，例如600607的1993-12-23日的交易量为nan
                    df = df.fillna(0)

                    cols_table = self.get_column_names(table_name)
                    for col in cols_table:  # 补齐数据库中有的列但是df中没有的列
                        if col not in df.columns:
                            df[col] = None

                    if isinstance(primary_type, DATE):
                        df['new'] = df.apply(
                            func=lambda r: False if pd.to_datetime(
                                r[primary_key]).date() in items_exist else True,
                            axis=1)
                    elif isinstance(primary_type, DATETIME):
                        df['new'] = df.apply(
                            func=lambda r: False if pd.to_datetime(
                                r[primary_key]) in items_exist else True,
                            axis=1)
                    else:
                        df['new'] = df.apply(
                            func=lambda r: False if r[primary_key] in items_exist else True, axis=1)
                    # 2.0版本需要将Timestamp转换为str类型，pandas中的字符串在插入数据库时，会由str强制转换成DATE类型
                    if len(df) > 0:
                        row1: pd.Series = df.iloc[0]  # df中的第一个记录
                        for title, value in row1.items():
                            if isinstance(value, pd.Timestamp):
                                df[title] = df[title].astype(str)

                    df_new = df[df['new']]
                    # 已存在的数据，视情况忽略或更新，0==False，写False会报PEP == False的警告
                    df_old = df[df['new'] == 0]
                    # ---------------------- 判断需要插入的和需要更新的数据 ---------------------------

                    # ---------------------- 判断是否需要执行插入操作 ---------------------------------
                    if if_exists == "append":  # 存在的数据不操作，只添加不存在的数据
                        if len(df_new) == 0:  # 如果只是追加不存在的数据，且插入的数据全部存在，则无须执行插入操作
                            # 此时将数据更新时间更新一下，因为也是验证过后了
                            self.update_update_time_of_table(
                                table_name)  # 最后更新数据库中数据的最后更新日期
                            return
                    # ---------------------- 判断是否需要执行插入操作 ---------------------------------

                    # 按数据库表的列顺序排列dataframe列，无论if_exists取replace还是append，都要插入新值
                    values_new = [dict(zip(cols_table, item))
                                  for _, item in df_new[cols_table].iterrows()]
                    try:
                        # 该方法速度仅次于直接执行sql语句，尽量使用该方法
                        session.execute(table.insert(), values_new)

                        if if_exists == "replace":  # 如果存在的数据需要替换
                            values_old = [dict(zip(cols_table, item))
                                          for _, item in df_old[cols_table].iterrows()]
                            session.execute(table.update(), values_old)
                    except:
                        traceback.print_exc()
                        logger.debug(
                            f"{table=}, values_new={slice_str(values_new)}")
                else:
                    logger.warning("未处理的数据插入")
            else:
                # 插入大数据量是特别慢，尤其是数据库已有重复记录时
                logger.warning(
                    f"df.to_sql方法创建的表{table_name}没有主键，建议先创建表后，再插入数据")
                df.to_sql(table_name, self.engine, if_exists=if_exists,
                          index=index)  # 该方法设置的数据库表没有主键
            self.update_update_time_of_table(table_name)  # 最后更新数据库中数据的最后更新日期
        except PendingRollbackError:
            try:
                session.rollback()  # 尝试回滚，但可能已经有一个回滚在进行中
            except Exception as rollback_exc:
                # 记录回滚失败的信息
                print(f"回滚失败: {rollback_exc}")
        finally:
            session.commit()
            session.close()

    def get_count_of_table(self, table_name: str):
        """
        获取表中数据的行数
        """
        session = self.SessionMaker()
        # 获取表中的记录数量
        table: Table = self.get_table(table_name)
        count = session.query(func.count(table.columns[0])).scalar()
        session.close()
        return count

    def get_last_record(self, table_name: str, refer_column: str, ascend=True):
        """
        获取表中最后一条数据
        :param table_name 表名
        :param refer_column 排序的列名
        :param ascend 是否升序排列，默认为真
        """
        table = self.get_table(table_name)
        session = self.SessionMaker()

        if ascend:  # table.id可以写成table.c.get('id')
            res = session.query(table).order_by(
                # 因为是first()，所有相当于排序取反
                desc(table.c.get(refer_column))).first()
        else:
            res = session.query(table).order_by(
                asc(table.c.get(refer_column))).first()
        session.close()
        return res

    def insert_item(self, table_name: str = None, values: list = None,
                    col_names: list = None, ignore=False,
                    replace=False, filter_warning=None):
        """
        向数据库中插入数据，这里传入的列名不要用反引号括起来，增

        values中的数据类型需要与表中每列的数据类型一致，数据库和python中数据类型对应如下：
        SqlAlchemy          python
        DATETIME            datetime.datetime/datetime.date
        VARCHAR             str

        cols_names和values两个列表一一对应。

        :param table_name: 表名
        :param values:
        :param col_names: 列名，如果是插入带有auto_increment属性的表数据，则必须指定列名，否则就需要指定auto_increment属性的字段的值
        :param ignore: 当插入数据重复时，是否忽略
        :param replace: 当插入数据重复时，是否替换，ignore和replace不能同时为True
        :param filter_warning: 过滤警告信息，[1062, 1265, 1366]，分别对应["Duplicate", "Data truncated", "Incorrect integer value"]
        """
        _: Table = self.get_table(table_name)
        session = self.SessionMaker()
        # ------------------ 该段语句在sqlite服务器上测试成功，但mysql5.6服务器上测试数据不更新 -------------------
        # if col_names is None:
        #     col_names = self.get_column_names(table_name)
        # paras = []
        # for col, val in zip(col_names, values):
        #     if isinstance(val, datetime.datetime) or isinstance(val, datetime.date):
        #         val = val.__repr__()
        #         paras.append(f"{col}={val}")
        #     else:
        #         paras.append(f"{col}='{val}'")
        # paras = ",".join(paras)
        # state = f"_.insert().values({paras})"
        # eval(state)
        # self.session.commit()
        # ------------------ 该段语句在sqlite服务器上测试成功，但mysql5.6服务器上测试数据不更新 -------------------

        # ------------------ mysql5.6以下语句测试成功 -------------------------------
        if replace:
            columns = ""
            self.insert_item()
        else:
            values = dict(zip(col_names, values))
            try:
                ins = _.insert(values=values)  # 低版本sqlalchemy语法1.4.39
                session.execute(ins)  # self.conn.execute()
                session.commit()
            except PendingRollbackError:
                try:
                    session.rollback()  # 尝试回滚，但可能已经有一个回滚在进行中
                except Exception as rollback_exc:
                    # 记录回滚失败的信息
                    print(f"回滚失败: {rollback_exc}")
                finally:
                    session.close()
                    return
            except TypeError:  # TypeError: TableClause.insert() got an unexpected keyword argument 'values'
                # ins = _.insert().values(values)  # 高版本sqlalchemy语法2.0.36
                stmt = (
                    insert(_).
                    values(values)
                )
                try:
                    # with self.engine.connect() as conn:
                    #     conn.execute(stmt)
                    #     conn.commit()
                    session.execute(stmt)
                    session.commit()
                except PendingRollbackError:
                    try:
                        session.rollback()  # 尝试回滚，但可能已经有一个回滚在进行中
                    except Exception as rollback_exc:
                        # 记录回滚失败的信息
                        print(f"回滚失败: {rollback_exc}")
                finally:
                    session.close()

    def has_table(self, table_name):
        """
        判断数据库中是否存在某个表
        """
        # return self.engine.has_table(table_name) # 该写法不兼容sqlalchemy 2.0版本
        # 2.0已测试，该方法不需要连接池中的连接资源，如果可以的话优先使用
        return inspect(self.engine).has_table(table_name)
        # with self.engine.connect() as conn:
        #     res = self.engine.dialect.has_table(conn, table_name)
        # return res  # 该写法兼容sqlalchemy1.4和2.0版本

    def exist_table(self, table_name):
        """
        同has_table()
        """
        return self.has_table(table_name)

    def get_table(self, table_name=None) -> Table | None:
        """
        获取表名为table_name的表对象，返回的是 Table()对象。
        如果不传入table_name，则返回数据库中的所有表，返回的是Table()对象的列表。
        """
        if table_name is None:
            return self.meta_data.tables.values()
        if table_name in self._table_cache:
            return self._table_cache[table_name]

        # 使用优化后的查询
        if not self.has_table(table_name):
            return None

        # 只刷新单个表而不是全部metadata
        self.meta_data.clear()
        self.meta_data.reflect(bind=self.engine, only=[table_name])
        table = self.meta_data.tables.get(table_name)
        if table is not None and isinstance(table, Table):
            self._table_cache[table_name] = table
        return table

    def get_pool_info(self):
        """
        获取连接池信息
        """
        return self.engine.pool.status()


if __name__ == '__main__':
    sql = SqlOperator(create_engine('mysql+pymysql://sges:sges@sges.yangke.site:3306/sges',
                                    poolclass=QueuePool,
                                    pool_size=5,
                                    max_overflow=10,
                                    pool_timeout=30,  # 连接池中没有可用连接时的等待时间，超过该时间还没有可用连接，则报错
                                    pool_recycle=3600,  # 连接池中的连接1小时后会回收
                                    pool_pre_ping=True,  # 每次连接请求时，先测试返回的连接是否还存活，不存活则从连接池中取下一个连接
                                    ))
    # _ = sql.exist_table('user')
    # sql.insert_item('user', ['杨可', 'yangkexn@tpri.com.cn', 'test'], col_names=['user_name', 'email', 'password'])
    _ = sql.exists_in_table('user', condition_dict={
                            'username': '杨可', 'password': 'test'})
    print(sql.get_pool_info())
    print(_)



# 量化分析中使用的数据库对应的类
import datetime
import os
import time
import traceback
import pandas as pd
import sys
from yangke.common.config import logger
from yangke.dataset.YKSqlalchemy import (
    SqlOperator, YkColumn, Engine, Double, QueuePool, String,
    BOOLEAN, DATE, create_engine, VARCHAR
)
from yangke.base import timeout
from sqlalchemy import text  # 新增导入
import toml
from pathlib import Path


class Storage:
    def __init__(self, kind: str, sql_user=None, sql_passwd=None, sql_ip=None, sql_port=3306,
                 sql_db='stocks', sqlite_db='stocks.db'):
        """
        初始化股票数据存储工具，该工具需要用户已经在数据库服务器上创建了相应的db或schema，注意mysql存储内容默认不区分大小写，需要将collation
        修改为utf8mb4_bin以支持区分大小写。

        该类只负责数据的存储，股票数据的获取在marruiData.py或tushareData.py中

        :param kind: mysql,sqlite 分别表示使用mysql数据库或sqlite数据库存储
        :param sql_user: mysql数据库的用户名，kind=mysql时必须传入
        :param sql_passwd: mysql数据库的密码，kind=mysql时必须传入
        :param sql_ip: mysql数据库的ip，kind=mysql时必须传入
        :param sql_port: mysql数据库的端口，kind=mysql时必须传入
        :param sql_db: mysql数据库的db，kind=mysql时必须传入
        :param sqlite_db: sqlite数据库文件，kind=sqlite时必须传入
        """
        self.kind: str = kind.lower()  # 可取值mysql、sqlite、字符串路径
        self.table_all_stocks_info = 'all_stocks_info'
        self.table_holiday = 'holiday'
        self.table_modify_time = 'modify_time'

        if self.kind.lower() == 'mysql':
            settings_path = Path.cwd() / "settings.toml"
            mysql_settings = {}
            try:
                with open(settings_path, "r", encoding="utf-8") as f:
                    settings = toml.load(f)
                    mysql_settings = settings.get('mysql', {})
            except FileNotFoundError:
                logger.info("未找到 settings.toml 文件，正在生成新文件...")
                default_mysql_settings = {
                    "user": "请填入 MySQL 用户名",
                    "passwd": "请填入 MySQL 密码",
                    "host": "请填入 MySQL IP 地址",
                    "port": sql_port,
                    "db": sql_db
                }
                settings = {"mysql": default_mysql_settings}
                with open(settings_path, "w", encoding="utf-8") as f:
                    toml.dump(settings, f)
                logger.info(f"已生成新的 settings.toml 文件，路径: {settings_path}")
                mysql_settings = default_mysql_settings
                logger.warning(f"请填写 settings.toml 文件中的缺失信息后重新运行，文件路径为：{settings_path}")
                exit(0)
            except toml.TomlDecodeError:
                logger.error(f"settings.toml 文件格式错误，mysql加载失败！")

            # 补充缺失的 MySQL 属性字段
            if not all([sql_user, sql_passwd, sql_ip]):
                sql_user = sql_user or mysql_settings.get('user')
                sql_passwd = sql_passwd or mysql_settings.get('passwd')
                sql_ip = sql_ip or mysql_settings.get('host')
                sql_port = sql_port or mysql_settings.get('port', 3306)
                sql_db = sql_db or mysql_settings.get('db', 'stocks')

                if not all([sql_user, sql_passwd, sql_ip]):
                    if 'user' not in mysql_settings:
                        mysql_settings['user'] = "请填入 MySQL 用户名"
                    if 'passwd' not in mysql_settings:
                        mysql_settings['passwd'] = "请填入 MySQL 密码"
                    if 'host' not in mysql_settings:
                        mysql_settings['host'] = "请填入 MySQL IP 地址"
                    if 'port' not in mysql_settings:
                        mysql_settings['port'] = sql_port
                    if 'db' not in mysql_settings:
                        mysql_settings['db'] = sql_db

                    settings = {"mysql": mysql_settings}
                    with open(settings_path, "w", encoding="utf-8") as f:
                        toml.dump(settings, f)
                    logger.info("已在 settings.toml 文件中补充缺失的 MySQL 属性字段，请进行填写。")
                    exit(0)

        if self.kind.lower() == 'sqlite':
            self.encoding = 'utf8'
            # sqlite数据库默认utf8编码，不需要指定，且连接字符串也不支持指定编码
            self.engine: Engine = create_engine(
                f'sqlite:///{sqlite_db}', echo=False)
        elif self.kind.lower() == 'mysql':
            self.encoding = 'utf8mb4'  # mysql数据库的utf8mb4编码兼容性更强
            self.error: str | None = None  # 新增：存储初始化错误信息

            try:
                # 新增：先连接到默认数据库（mysql）检查目标库是否存在
                temp_engine = create_engine(
                    f"mysql+pymysql://{sql_user}:{sql_passwd}@{sql_ip}:{sql_port}/mysql?charset={self.encoding}",
                    echo=False
                )
                with temp_engine.connect() as conn:
                    # 检查当前用户是否有创建数据库的权限
                    grants = conn.execute(
                        text("SHOW GRANTS FOR CURRENT_USER()")).fetchall()
                    has_create_privilege = any(
                        'CREATE' in grant[0] for grant in grants)
                    if not has_create_privilege:
                        self.error = "当前用户没有创建数据库的权限，请联系管理员授予 CREATE 权限（如：GRANT CREATE ON *.* TO '当前用户'@'当前主机'）"
                        logger.error(self.error)
                        self.engine = None  # 标记引擎不可用
                        return  # 提前终止初始化流程

                    # 检查数据库是否存在
                    check_db_sql = text(
                        f"SELECT SCHEMA_NAME FROM INFORMATION_SCHEMA.SCHEMATA WHERE SCHEMA_NAME = '{sql_db}'")
                    db_exists = conn.execute(check_db_sql).fetchone()

                    if not db_exists:
                        # 不存在则创建数据库（保留原有字符集和排序规则设置）
                        create_db_sql = text(
                            f"CREATE DATABASE {sql_db} CHARACTER SET {self.encoding} COLLATE utf8mb4_bin")
                        conn.execute(create_db_sql)
                        logger.info(
                            f"数据库 {sql_db} 不存在，已自动创建（字符集：{self.encoding}，排序规则：utf8mb4_bin）")
            except Exception as e:
                self.error = f"当前用户可能没有创建数据库的权限，需确保{sql_db}存在！"

            # 创建目标数据库引擎
            self.engine: Engine = create_engine(
                f"mysql+pymysql://{sql_user}:{sql_passwd}@{sql_ip}:{sql_port}/{sql_db}?charset={self.encoding}",
                echo=False,
                poolclass=QueuePool,
                pool_size=5,
                max_overflow=10,
                pool_timeout=30,  # 连接池中没有可用连接时的等待时间，超过该时间还没有可用连接，则报错
                pool_recycle=3600,  # 连接池中的连接1小时后会回收
                pool_pre_ping=True  # 每次连接请求时，先测试返回的连接是否还存活，不存活则从连接池中取下一个连接
            )
        else:
            logger.error("暂不支持除了'mysql'和'sqlite'之外的其他存储方式！")
            self.engine = None

        self.sql: SqlOperator = SqlOperator(self.engine) # 放到最后，则确保初始化的几个表已经存在
        self.daily_all_data: list[dict] | None = None  # 所有股票最后一天的日线数据
        self.init()


    def init(self):
        if not self.sql.has_table(self.table_holiday):
            self.sql.create_table(self.table_holiday,
                                  columns=[
                                      YkColumn(
                                          'date', DATE, nullable=False, primary_key=True),  # 日期
                                      YkColumn('open', BOOLEAN,
                                               nullable=False)  # 是否开盘
                                  ])
        if not self.sql.has_table(self.table_all_stocks_info):
            self.sql.create_table(table_name=self.table_all_stocks_info,
                                  columns=[  # mysql默认不区分大小写，这里设置collation为区分大小写
                                      YkColumn('symbol', VARCHAR(
                                          10), nullable=False, primary_key=True),  # 000001
                                      YkColumn('name', String(20)),  # 平安银行
                                      YkColumn('exchange', String(10)),  # SZ
                                      YkColumn('market', String(10)),  # SZ
                                  ])
        if not self.sql.has_table(self.table_modify_time):
            self.sql.create_all_base_table()

    @timeout(10)
    def create_stock_table_day(self, code):
        """
        为代码为code的股票创建日线表
        :param code: 股票代码，如"600001"，对应的数据库表名为"daily600001"
        """
        table_name = f"daily{code}"
        if not self.sql.has_table(table_name):
            self.sql.create_table(table_name=table_name,
                                  columns=[
                                      YkColumn('trade_date', DATE(),
                                               primary_key=True),
                                      # 在Ubuntu系统测试，Mysql的Float类型只有6位有效数字
                                      YkColumn('open', Double()),
                                      YkColumn('high', Double()),
                                      YkColumn('low', Double()),
                                      YkColumn('close', Double()),
                                      YkColumn('vol', Double()),
                                      YkColumn('amount', Double()),
                                      YkColumn('换手率', Double()),
                                      YkColumn('市值', Double(), nullable=True),
                                      YkColumn('流通市值', Double(),
                                               nullable=True),
                                      YkColumn('市盈率', Double(), nullable=True),
                                      YkColumn('市净率', Double(), nullable=True),
                                  ])

    def update_all_stocks_table(self, df: pd.DataFrame | None):
        self.sql.insert_dataframe(self.table_all_stocks_info, df)

    def get_all_stock_basic_info(self) -> pd.DataFrame:
        all_stocks = self.sql.select_item(
            table_name=self.table_all_stocks_info, result_type='json')
        return pd.DataFrame(all_stocks)

    def need_update_all_stocks_table(self):
        """
        检查所有股票信息表是否需要更新。

        :return: 如果需要更新返回True，否则返回False
        """
        last_time = self.sql.get_update_time_of_table(
            self.table_all_stocks_info)
        return self.need_update(last_time)

    def get_modify_time_all(self):
        """
        获取所有表格的最后更新时间
        """
        items = self.sql.select_item(table_name=self.table_modify_time)
        return items

    def get_need_single_day_update_stocks(self):
        """
        获取只需要补充最新一天数据的股票列表
        """
        ...

    def need_update_holiday_table(self, date: datetime.date = None):
        """
        是否需要更新假期表，如果能查询到指定日期的数据项，则不需要更新，否则需要更新
        """
        if date is None:
            date = datetime.date.today()  # 不能使用datetime.datetime.today()否则时分秒对不上
        item = self.sql.select_item(self.table_holiday, {"date": date})
        if item is None:
            return True
        else:
            return False

    @timeout(10)
    def need_update_daily(self, symbol: str):
        """
        检查指定股票的日线数据是否需要更新。

        :param symbol: 股票代码，如"600001"
        :return: 如果需要更新返回True，否则返回False
        """
        table_name = f"daily{symbol}"
        last_time = self.sql.get_update_time_of_table(table_name)
        return self.need_update(last_time)

    @timeout(100)
    def get_last_date_of_daily_table(self, symbol: str):
        """
        获取日线数据表中的最后一条数据的日期
        """
        table_name = f"daily{symbol}"
        res = self.sql.get_last_record(table_name, 'trade_date', True)
        return res if res is None else res.trade_date

    @timeout(240)
    def update_daily(self, symbol: str, daily_data: pd.DataFrame):
        """
        更新指定股票的日线数据
        """
        table_name = f"daily{symbol}"
        if not self.sql.has_table(table_name):
            self.create_stock_table_day(symbol)
        self.sql.insert_dataframe(table_name, daily_data)

    def update_daily_all(self, daily_all_data: str):
        """
        更新所有股票的日线数据
        """
        self.daily_all_data = daily_all_data

    def update_holiday_table(self, holiday: pd.DataFrame):
        """
        更新假期表的数据。

        :param holiday: 包含假期信息的DataFrame对象
        """
        self.sql.insert_dataframe(self.table_holiday, holiday)

    @timeout(200)
    def is_holiday(self, day_datetime: datetime.date = datetime.date.today()):
        """
        判断今天是不是休市，只适用于国内股市，不包括港股

        :param day_datetime: 给定日期的datetime类型
        :return: True 或 False
        """
        day_str = day_datetime.strftime("%Y%m%d")
        if (day_datetime.weekday() == 5) or (day_datetime.weekday() == 6):  # 周六和周日肯定是休市的，这里的5、6分别代表周六和周日
            return True
        date_str = day_str
        is_open = self.sql.select_in_table(table_name=self.table_holiday, condition_dict={"date": day_datetime},
                                           result_col=['open'])
        is_open = bool(is_open)  # 1：True，0：False
        if is_open is None:  # 未查找到相关记录
            year = day_datetime.year
            # try:
            #     holiday_data: pd.DataFrame = self.api.query('trade_cal', start_date=f'{year}0101',
            #                                                 end_date=f'{year}1231')  # tushare该接口现在收费了，不让使用
            #     holiday_data.rename(
            #         columns={"cal_date": "calendarDate", "is_open": "isOpen"}, inplace=True)
            # except:
            #     traceback.print_exc()
            #     logger.debug('似乎没有tushare的数据获取权限，尝试使用baostock获取数据')
            #     import baostock as bs
            #     bs.login()
            #     holiday_data: pd.DataFrame = bs.query_trade_dates(start_date=f'{year}-01-01',
            #                                                       end_date=f'{year}-12-31').get_data()
            #     bs.logout()
            #     holiday_data.rename(columns={
            #                         "calendar_date": "calendarDate", "is_trading_day": "isOpen"}, inplace=True)
            # 日期必须转换为datetime才能与数据库中的DATE列匹配
            # holiday_data["calendarDate"] = pd.to_datetime(
            #     holiday_data["calendarDate"])
            # holiday_data["isOpen"] = holiday_data["isOpen"].apply(
            #     lambda x: True if int(x) == 1 else False)

            # # to_sql()语句存储时，使用dtype参数指定dataframe到数据库的对应类型
            # # holiday_data[["calendarDate", "isOpen"]].to_sql(self.假期表, con=self.engine, if_exists='append',
            # #                                                 index=False,
            # #                                                 dtype={"calendarDate": DATE, "isOpen": BOOLEAN})
            # self.sql.insert_dataframe(self.table_holiday, holiday_data, if_exists='append', index=False,
            #                           dtype={"calendarDate": DATE, "isOpen": BOOLEAN})

            # if holiday_data[holiday_data['calendarDate'] == date_str]['isOpen'].item():
            #     return False
            # else:
            #     return True
        else:
            return not is_open  # python会自动将0转换成False，1转换成True

    def need_update(self, data_file: str or datetime.datetime):
        """
        检查数据文件是否需要更新；如果传入的是时间，则检查该时间之后是否有股票数据更新。
        当上一个工作日4点以后更新过，且今天时间不到下午4点，则无需更新

        :param data_file:
        :return:
        """
        if isinstance(data_file, datetime.datetime):
            last_change_time = data_file
            last_change_date_str = last_change_time.strftime("%Y-%m-%d")
        elif data_file is None:  # 如果文件为空
            return True
        elif os.path.exists(data_file):  # 如果文件存在，判断文件是否需要更新
            last_change_time = os.stat(data_file).st_mtime
            last_change_time_str = time.strftime(
                "%Y-%m-%d %H:%M:%S", time.localtime(last_change_time))
            last_change_time = datetime.datetime.strptime(
                last_change_time_str, "%Y-%m-%d %H:%M:%S")
            last_change_date_str = last_change_time.strftime("%Y-%m-%d")
        else:  # 文件不存在，则需要更新
            return True

        now_time = datetime.datetime.now()  # 当前时间
        now_time_date_str = now_time.strftime("%Y-%m-%d")  # 用来和最后修改日期作比较

        if last_change_date_str == now_time_date_str:  # 如果是今天修改的
            if last_change_time.hour > 15:  # 最后修改日期是当天15点以后，则不用更新数据
                logger.debug(f"数据文件{data_file}已经最新，无需更新！")
                return False  # 无需更新数据
            else:
                if now_time.hour < 16:  # 如果今天修改过且今天还没到16:00，也无须更新
                    return False
                # 如果今天时holiday
                if self.is_holiday(datetime.date.today()):  # 如果今天时休息日，今天更新过也不用更新
                    logger.debug(f"数据文件{data_file}已经最新，无需更新！")
                    return False
                else:
                    return True
        else:
            last_working_day_str = str(
                self.get_previous_working_day(include_time=True)) + ' 16:00:00'
            last_working_datetime = datetime.datetime.strptime(
                last_working_day_str, '%Y-%m-%d %H:%M:%S')
            if last_change_time > last_working_datetime:
                # 如果最后修改日期在上一个工作日15点之后
                if isinstance(data_file, str):
                    logger.debug(
                        f"数据文件{data_file}已经最新，无需更新！{last_working_datetime=}")
                else:
                    logger.debug(
                        f"数据更新于{data_file}，无需更新！{last_working_datetime=}")
                return False  # 无需更新数据
            else:
                return True

    def get_previous_working_day(self, last_date=None, include_time=False) -> datetime.date:
        """
        获得指定日期的上一个交易日，如果不传入参数，则获得最近的一个交易日，包含今天。

        :param last_date: 指定的日期
        :param include_time: 是否判断时间，如果includeTime=True，则会进一步判断今天的时间，如果时间在下午4:00之前，则不包括今天，因为当天的股票数据还没整理出来
        :return: 最近一个交易日日期
        """
        if last_date is None:  # 不能直接在定义函数时给初值，否则初值将在函数第一次运行时就确定，后续无论运行多少次，都不会改变
            last_date = datetime.date.today()
        if last_date > datetime.date.today():
            logger.warning("因为未来的股票数据不存在，不能获得将来日期的前一个工作日！")
        if self.is_holiday(last_date):
            one_day = datetime.timedelta(days=1)
            pre_day = last_date - one_day
            while self.is_holiday(pre_day):
                pre_day = pre_day - one_day
            return pre_day
        else:
            if include_time and last_date == datetime.date.today():  # 需要判断时间，且last_date是今天
                dt = datetime.datetime.now()
                hour = dt.hour
                if hour < 16:  # 额外对时间进行判断
                    return self.get_working_day_before_day(datetime.date.today())
                else:
                    return last_date
            return last_date

    def get_working_day_before_day(self, day_datetime: None | datetime.date,
                                   day_num: int = 1) -> datetime.date:
        """
        获得指定日期的前一个交易日或后一个交易日，需要注意的是，该函数不对指定日期是否为工作日进行判断，
        因此，获得最近一个交易日，不能使用get_working_day_before_day(date.today(),1)
        :param day_datetime: datetime类型，默认为今天
        :param day_num: 可取正整数和-1，前day_num个交易日，取-1时,表示指定日期的后一个交易日，不可以取其他负值
        :return: date_datetime】,-----返回对应交易日的datetime类型数据
        """
        if day_datetime is None:
            day_datetime = datetime.date.today()
        if day_num < -1:
            logger.error("day_num不能取-1以外的负值（{}）".format(day_num))
            sys.exit()
        if day_num == 1 or day_num == -1:
            one_day = datetime.timedelta(days=1)
            if day_num == -1:
                pre_day = day_datetime + one_day
            else:
                pre_day = day_datetime - one_day
            while self.is_holiday(pre_day):
                if day_num == -1:
                    pre_day = pre_day + one_day
                else:
                    pre_day = pre_day - one_day
            return pre_day
        else:  # day_num为大于1的整数
            day_datetime = self.get_working_day_before_day(
                day_datetime, 1)  # 计算上一个交易日
            # 返回上一个交易日的上day_num-1个交易日
            return self.get_working_day_before_day(day_datetime, day_num - 1)

    @timeout(seconds=100)
    def downloadHolidays(self, args):
        res = self.sql.select_item(self.table_holiday, result_type="json")
        # 判断数据库中是否存在当天的数据，不存在则更新

        # res = [{'date': item['date'].strftime('%Y-%m-%d'), 'open': item['open']} for item in res]
        return res

    @timeout(seconds=100)
    def downloadStocksTotalInfo(self, args):
        res = self.sql.select_item(
            self.table_all_stocks_info, result_type="json")
        return res

    @timeout(seconds=100)
    def downloadStocksDailySingle(self, args: dict):
        symbol = args.get("symbol", None)
        if symbol is None:
            return None
        res = self.sql.select_item(f"daily{symbol}", result_type="json")
        return res

    def downloadStocksDailyAll(self, args: dict):
        return {"info": "暂不支持"} | args

    @timeout(seconds=100)
    def downloadLastDay(self, args):
        """
        下载最后一天所有股票的日线数据
        """
        return self.daily_all_data

    def start_rest_service(self, rest_port=5002, single_thread=False, daemon=True):
        """
        开启restful服务，等待客户端连接
        """
        from yangke.web.flaskserver import start_server_app

        def deal(args):
            try:
                # 因为下方use_action=True，所以这里的action必然有值，避免eval函数出错
                action = args.get('action')
                result = eval(f"self.{action}(args)", {}, {
                              "self": self, "args": args})
                return result
            except TimeoutError:
                logger.warning(f"函数执行超时")
                return {"success": False,
                        "info": "执行deal时错误，函数执行超时"}
            except:
                traceback.print_exc()
                return {"success": False,
                        "info": "执行deal时错误"}

        start_server_app(deal=deal, use_action=True,
                         allow_action=['downloadHolidays', 'downloadStocksDailyAll', 'downloadStocksDailySingle',
                                       'downloadStocksTotalInfo', 'downloadLastDay'],
                         host='0.0.0.0',
                         port=rest_port,
                         example_url=[
                             f'http://localhost:{rest_port}/?Action=downloadHolidays'],
                         single_thread=single_thread)


if __name__ == '__main__':
    storage = Storage('mysql')



import datetime
import time
import toml
from pathlib import Path
import pandas as pd
import requests
from yangke.base import timeout
from yangke.common.config import logger
import baostock as bs
import tushare as ts


class StockData:
    def __init__(self, license=None):
        """
        获取股票数据的类，该类只负责数据获取，数据的存储在storage.py中
        """
        super().__init__()
        # 从settings.toml加载license（当入参为None时）
        if license is None:
            settings_path = Path.cwd() / "settings.toml"
            try:
                with open(settings_path, "r", encoding="utf-8") as f:
                    settings = toml.load(f)
                self.license = settings.get("license")
                self.tushare_token = settings.get(
                    "tushare_token")  # 加载tushare token
                if not self.license:
                    logger.info("settings.toml中未找到有效的license字段，正在添加...")
                    settings["license"] = "请在这里填入您的【麦蕊智数】许可证密钥"
                    # 添加tushare token字段
                    settings["tushare_token"] = "请在这里填入您的tushare token，不填写则不能使用tushare的接口"
                    with open(settings_path, "w", encoding="utf-8") as f:
                        toml.dump(settings, f)
                    logger.info(f"已在settings.toml文件中添加license字段，请进行填写。")
                    raise ValueError("请在settings.toml文件中填入有效的license字段")
            except FileNotFoundError:
                logger.info("未找到settings.toml文件，正在生成新文件...")
                default_settings = {
                    "license": "请在这里填入您的【麦蕊智数】许可证密钥"
                }
                try:
                    with open(settings_path, "w", encoding="utf-8") as f:
                        toml.dump(default_settings, f)
                    logger.info(f"已生成新的settings.toml文件，路径: {settings_path}")
                    raise ValueError("请在settings.toml文件中填入有效的license字段")
                except Exception as e:
                    logger.error(f"生成settings.toml文件时出错: {str(e)}")
                    raise
            except toml.TomlDecodeError:
                raise ValueError("settings.toml文件格式错误")
            except ValueError as e:
                logger.error(f"请填写settings.toml文件中的license字段！")
                exit(0)

        else:
            self.license = license
        self.pre_url = 'https://y.mairui.club'

    def get_market(self, symbol):
        """
        获取股票所属的板块
        """
        symbol = str(symbol)
        if symbol.startswith('60'):
            return '主板'
        elif symbol.startswith('00'):
            return "主板"
        elif symbol.startswith('30'):
            return '创业板'
        elif symbol.startswith('68'):
            return '科创板'
        elif symbol.startswith('82'):
            return '优先股'
        elif symbol.startswith('83'):
            return '普通股'
        elif symbol.startswith('87'):
            return '普通股'
        elif symbol.startswith('4'):
            return '北交所'
        return None

    def get_type_szsh(self, symbol):
        """
        判断股票所属的交易所，代码来自网络

        沪市股票包含上证主板和科创板和B股：沪市主板股票代码是60开头、科创板股票代码是688开头、B股代码900开头。
        深市股票包含主板、中小板、创业板和B股：深市主板股票代码是000开头、中小板股票代码002开头、创业板300开头、B股代码200开头
        """
        symbol = str(symbol)
        if symbol.find('60', 0, 3) == 0:
            gp_type = 'sh'
        elif symbol.find('688', 0, 4) == 0:
            gp_type = 'sh'
        elif symbol.find('900', 0, 4) == 0:
            gp_type = 'sh'
        elif symbol.find('00', 0, 3) == 0:
            gp_type = 'sz'
        elif symbol.find('300', 0, 4) == 0:
            gp_type = 'sz'
        elif symbol.find('200', 0, 4) == 0:
            gp_type = 'sz'
        return gp_type

    def get_all_stock_basic_info(self) -> pd.DataFrame | None:
        """
        获得上海和深圳证券交易所目前上市的所有股票代码
        """
        url = f'{self.pre_url}/hslt/list/{self.license}'
        try:
            response = requests.get(url, timeout=10)
            response.raise_for_status()  # 主动触发HTTP错误异常
            if response.status_code == 200:
                df = pd.DataFrame(response.json())
                df.rename(columns={"dm": "symbol", "mc": "name",
                                   "jys": "exchange"}, inplace=True)
                # 统一股票代码格式为6位数字
                df['symbol'] = df['symbol'].apply(
                    lambda x: x[:6] if '.' in x else x)
                # 将exchange列统一改为小写
                df['exchange'] = df['exchange'].str.lower()
                # 移除lambda函数，直接使用函数引用
                df['market'] = df['symbol'].apply(self.get_market)
                df = df.sort_values(by='symbol', ascending=True)
                df = df.reset_index().drop(columns='index')
                return df
            else:
                logger.debug(f"获取数据失败，检查网络或证书，尝试连接{url}")
                return None
        except requests.exceptions.RequestException as e:
            logger.error(f"获取股票基本信息失败: {str(e)}")
            logger.debug(f"请求URL: {url}")
            logger.debug(f"响应状态码: {getattr(response, 'status_code', '无响应')}")
            return None

    @timeout(240)
    def get_daily(self, symbol, end_date: datetime.date = None):
        """
        获取指定股票的日线数据
        """
        # "https://y.mairui.club/zs/hfsjy/000001/dn/lincense"
        # 根据《股票列表》得到的股票代码和分时级别获取历史交易数据，交易时间从远到近排序。目前 分时级别 支持5分钟、15分钟、30分钟、60分钟、日
        # 周月年级别（包括前后复权），对应的值分别是 5m（5分钟）、15m（15分钟）、30m（30分钟）、60m（60分钟）、dn(日线未复权)、dq（日线前复权）
        # 、dh（日线后复权）、wn(周线未复权)、wq（周线前复权）、wh（周线后复权）、mn(月线未复权)、mq（月线前复权）、mh（月线后复权）、
        # yn(年线未复权)、yq（年线前复权）、yh（年线后复权）
        url = f'{self.pre_url}/zs/hfsjy/{symbol}/dn/{self.license}'
        try:
            response = requests.get(url, timeout=60)
            if response.status_code == 404:
                logger.error(f"股票{symbol}日线数据未找到, URL: {url}")
                if symbol == "000508":
                    logger.debug("该股已退市")
                return None
            elif response.status_code == 200:
                res = pd.DataFrame(response.json())
                if res.empty:
                    logger.warning(f"股票{symbol}日线数据为空, URL: {url}")
                    return res

                res.rename(columns={"d": "trade_date",
                                    "o": "open", "h": "high",
                                    "l": "low", "c": "close",
                                    "v": "vol", "e": "amount",
                                    "hs": "换手率",
                                    "sz": "市值",  # 总市值
                                    "lt": "流通市值",
                                    "pe": "市盈率",
                                    "sjl": "市净率"
                                    },
                           inplace=True)

                if end_date is None:
                    return res
                # 如果res中的trade_date的最后一天不是end_date，则下载end_date到res中的trade_date的最后一天之间的数据
                if end_date is not None:
                    if isinstance(end_date, datetime.date):
                        end_date = end_date.strftime("%Y-%m-%d")
                    logger.debug(
                        f"end_date: {end_date}, res_last_date: {res.iloc[-1]['trade_date']}")
                    # 检查是否需要补充数据
                    if not res.empty and res.iloc[-1]["trade_date"] != end_date:
                        try:
                            start_date = datetime.datetime.strptime(
                                res.iloc[-1]["trade_date"], "%Y-%m-%d").date() + datetime.timedelta(days=1)
                            end_date_dt = datetime.datetime.strptime(
                                end_date, "%Y-%m-%d").date()
                            if start_date >= end_date_dt:
                                logger.warning(
                                    f"开始日期{start_date}不能大于等于结束日期{end_date}")
                                return res

                            append_data = self.get_daily_range(
                                symbol, start_date, end_date_dt)
                            if append_data is not None and not append_data.empty:
                                # 将字符串日期转换为日期对象进行比较
                                append_data_dates = pd.to_datetime(
                                    append_data["trade_date"])
                                res_dates = pd.to_datetime(res["trade_date"])
                                append_data = append_data[~append_data_dates.isin(
                                    res_dates)]
                                res = pd.concat(
                                    [res, append_data], ignore_index=True)
                                # 按trade_date排序
                                res["trade_date"] = pd.to_datetime(
                                    res["trade_date"])
                                res = res.sort_values("trade_date")
                                res["trade_date"] = res["trade_date"].dt.strftime(
                                    "%Y-%m-%d")
                        except Exception as e:
                            logger.warning(f"补充数据时出错: {str(e)}")
                return res
            return None
        except requests.exceptions.SSLError:
            logger.debug("请求数据发生SSLError，间隔1分钟后重新尝试")
            time.sleep(60)  # 间隔60s后重试
            return self.get_daily(symbol, end_date)

    def get_daily_sh_index(self, end_date=None):
        """
        获取上证指数的日线
        """
        symbol = 'sh000001'
        return self.get_daily(symbol, end_date)

    def get_daily_sz_index(self, end_date=None):
        """
        获取深证指数的日线
        """
        symbol = ''
        return self.get_daily(symbol, end_date)

    @timeout(100)
    def get_single_day_data(self, symbol):
        """
        获取指定股票的日线数据
        http://api1.mairui.club/hsrl/ssjy/股票代码(如000001)/您的licence
        """
        url = f'{self.pre_url}/hsrl/ssjy/{symbol}/{self.license}'
        # 添加超时参数，避免程序无限期挂起
        response = requests.get(url, timeout=10)
        res = pd.DataFrame([response.json()])
        res.rename(columns={"t": "trade_date",
                            "o": "open", "h": "high",
                            "l": "low", "p": "close",
                            "v": "vol", "cje": "amount",
                            "hs": "换手率",
                            "sz": "市值",  # 总市值
                            "lt": "流通市值",
                            "pe": "市盈率",
                            "sjl": "市净率"
                            },
                   inplace=True)
        return res

    def get_daily_all(self, try_times=1, need_json=False):
        """
        获取所有股票当天的交易数据
        https://y.mairui.club/hsrl/ssjy/all/lincense
        http://a.mairui.club/hsrl/ssjy/all/lincense  该地址已测试
        """
        url = f'http://a.mairui.club/hsrl/ssjy/all/{self.license}'
        try:
            response = requests.get(url, timeout=40)
            response.raise_for_status()  # 主动触发HTTP错误异常，会被requests.exceptions.RequestException捕获

            try:
                json_res = response.json()
                res = pd.DataFrame(json_res)
                res.rename(columns={
                    "t": "trade_date",
                    "o": "open", "h": "high",
                    "l": "low", "p": "close",
                    "v": "vol", "cje": "amount",
                    "hs": "换手率",
                    "sz": "市值",
                    "lt": "流通市值",
                    "pe": "市盈率",
                    "sjl": "市净率",
                    "dm": "symbol"
                }, inplace=True)

                if need_json:
                    return json_res, res
                return res

            except requests.exceptions.JSONDecodeError:
                logger.error(f"JSON解析失败: {response.text}")
                if try_times < 3:
                    time.sleep(70)
                    return self.get_daily_all(try_times + 1, need_json=need_json)
                return None

        except requests.exceptions.RequestException as e:
            logger.error(f"请求失败: {str(e)}")
            if try_times < 3:
                time.sleep(70)
                return self.get_daily_all(try_times + 1, need_json=need_json)
            return None

    @timeout(100)
    def get_daily_range(self, symbol, start_date: datetime.date, end_date: datetime.date, method='default'):
        """
        下载指定股票的指定日期范围的数据，不复权
        https://y.mairui.club/hsstock/history/000001.SZ/d/n/{license}?st=20250501&et=20250516
        分时级别支持1分钟、5分钟、15分钟、30分钟、60分钟、日线、周线、月线、年线，对应的请求参数分别为1、5、15、30、60、d、w、m、y
        除权方式有不复权、前复权、后复权、等比前复权、等比后复权，对应的参数分别为n、f、b、fr、br

        @param symbol: 股票代码
        @param start_date: 开始日期(格式:YYYYMMDD)
        @param end_date: 结束日期(格式:YYYYMMDD)
        @return: 包含历史数据的DataFrame，列名与get_daily方法一致
        """
        if method == 'default':
            if "." in symbol:
                symbol = symbol.upper()
            else:
                symbol = f"{symbol}.{self.get_type_szsh(symbol).upper()}"

            start_str = start_date.strftime("%Y%m%d")
            end_str = end_date.strftime("%Y%m%d")
            url = f'{self.pre_url}/hsstock/history/{symbol}/d/n/{self.license}?st={start_str}&et={end_str}'
            try:
                response = requests.get(url, timeout=60)
                if response.status_code == 404:
                    logger.error(
                        f"股票{symbol}在{start_date}至{end_date}范围内的数据未找到, URL: {url}")
                    return self.get_daily_range(symbol, start_date, end_date, method='baostock')
                elif response.status_code == 200:
                    res = pd.DataFrame(response.json())
                    res.rename(columns={
                        "t": "trade_date",
                        "o": "open", "h": "high",
                        "l": "low", "c": "close",
                        "v": "vol", "a": "amount",
                    }, inplace=True)
                    # 将trade_date转换为YYYY-MM-DD格式
                    res["trade_date"] = pd.to_datetime(
                        res["trade_date"]).dt.strftime("%Y-%m-%d")
                    res = res[["trade_date", "open", "high",
                               "low", "close", "vol", "amount"]]
                    return res
                return self.get_daily_range(symbol, start_date, end_date, method='baostock')
            except requests.exceptions.RequestException as e:
                logger.error(f"获取股票{symbol}历史数据失败，将从baostock重试: {str(e)}")
                logger.debug(f"请求URL: {url}")
                return self.get_daily_range(symbol, start_date, end_date, method='baostock')
        elif method == 'baostock':
            logger.debug(f"baostock获取数据中...")
            if "." in symbol and symbol.index('.') == 6:
                _1, _2 = symbol.split('.')
                symbol = f"{_2}.{_1}".lower()
            elif "." in symbol and symbol.index('.') == 2:
                symbol = symbol.lower()
            else:
                symbol = f"{self.get_type_szsh(symbol)}.{symbol}".lower()

            try:
                # 登录系统
                lg = bs.login()
                if lg.error_code != '0':
                    logger.error(f"baostock登录失败: {lg.error_msg}")
                    return self.get_daily_range(symbol, start_date, end_date, method='tushare')

                # 获取历史K线数据
                rs = bs.query_history_k_data_plus(
                    symbol,
                    "date,code,open,high,low,close,volume,amount,adjustflag",
                    start_date=start_date.strftime("%Y-%m-%d"),
                    end_date=end_date.strftime("%Y-%m-%d"),
                    frequency="d",  # d表示日线
                    adjustflag="3"  # 3表示不复权
                )

                if rs.error_code != '0':
                    logger.error(f"获取股票{symbol}历史数据失败: {rs.error_msg}")
                    bs.logout()
                    return self.get_daily_range(symbol, start_date, end_date, method='tushare')

                # 处理结果集
                data_list = []
                while (rs.error_code == '0') & rs.next():
                    data_list.append(rs.get_row_data())

                result = pd.DataFrame(data_list, columns=rs.fields)
                bs.logout()

                # 重命名列以保持与其他方法一致
                result.rename(columns={
                    "date": "trade_date",
                    "volume": "vol",
                    "amount": "amount"
                }, inplace=True)
                # 将成家量单位从“股”转换为“手”
                result['vol'] = result['vol']/100

                return result[["trade_date", "open", "high", "low", "close", "vol", "amount"]]
            except Exception as e:
                logger.error(f"baostock获取数据异常: {str(e)}")
                try:
                    bs.logout()
                except:
                    pass
                return self.get_daily_range(symbol, start_date, end_date, method='tushare')
        elif method == 'tushare':
            try:
                # 初始化pro接口
                pro = ts.pro_api(self.tushare_token)  # 使用类中已有的license

                # 构建股票代码
                if "." in symbol and symbol.index('.') == 6:
                    symbol = symbol.upper()
                elif "." in symbol and symbol.index('.') == 2:
                    _1, _2 = symbol.split('.')
                    symbol = f"{_2}.{_1}".upper()
                else:
                    symbol = f"{symbol}.{self.get_type_szsh(symbol).upper()}"

                # 获取数据
                df = pro.daily(
                    ts_code=symbol,
                    start_date=start_date.strftime("%Y%m%d"),
                    end_date=end_date.strftime("%Y%m%d"),
                    fields=[
                        "trade_date",
                        "open",
                        "high",
                        "low",
                        "close",
                        "vol",
                        "amount"
                    ]
                )

                if df.empty:
                    logger.warning(
                        f"tushare未找到{symbol}在{start_date}至{end_date}范围内的数据")
                    return None

                # 重命名列以保持与其他方法一致
                df.rename(columns={
                    "vol": "vol",
                    "amount": "amount"
                }, inplace=True)

                # 确保trade_date列是日期类型
                df["trade_date"] = pd.to_datetime(df["trade_date"])
                df.sort_values("trade_date", inplace=True)
                # 格式化日期为YYYY-MM-DD
                df["trade_date"] = pd.to_datetime(
                    df["trade_date"]).dt.strftime("%Y-%m-%d")

                return df[["trade_date", "open", "high", "low", "close", "vol", "amount"]]

            except Exception as e:
                logger.error(f"tushare获取数据异常: {str(e)}")
                return None
        return None


if __name__ == '__main__':
    sd = StockData()
    sd.get_all_stock_basic_info()



"""
假设3台电脑，其中：
电脑A，云服务器，具有公网IP182.43.65.44
电脑B，内网电脑，上面的20010端口提供了web服务
电脑C，其余可以连接互联网的电脑

如果需要将电脑B上的服务提供给公网，使得电脑C可以访问，则需要在电脑B上执行如下命令：
ssh -R 0.0.0.0:5000:127.0.0.1:20010 root@182.43.65.44
root@182.43.65.44是用root用户登陆电脑A，并将电脑B的20010端口映射到182.43.65.44的5000端口，如果一切正常，则可以在公网中通过182.43.65.44:5000访问内网提供的服务了。

如果执行后，公网仍不能访问，则需要修改云服务器(即电脑A)中的SSH设置。
vim /etc/ssh/sshd_config

设置以下参数，允许远程端口转发。
GatewayPorts yes

秘钥登陆设置的方法。
 1.首先生成密钥对，在电脑B上执行：
    ssh-keygen
 然后一致回车，知道生成id_isa和id_isa.pub文件

 2.将公钥上传到云服务器A上，分两种情况：
 如果电脑B是windows系统，需要在powershell中执行以下命令：
    type $env:USERPROFILE\.ssh\id_rsa.pub | ssh root@182.43.65.44 "cat >> .ssh/authorized_keys"
 其中，root@182.43.65.44表示root是远程登陆的用户名，182.43.65.44是远程登陆的服务器，即电脑A的ip地址。
 如果电脑B是linux系统，则直接执行以下命令：
    ssh-copy-id -i ~/.ssh/id_rsa.pub root@182.43.65.44
 即可。

如上设置后，可能还需要解决ssh保持活跃的设置


"""
from yangke.core import runCMD


def check_ssh_tunnel_exist(remote_host, remote_port):
    cmd = f"powershell wget https://{remote_host}:{remote_port}"
    msg, err = runCMD(cmd, charset='utf8', timeout=5)
    if "无法连接到远程服务器" in err:
        # 说明端口未进行隧道转发
        return False
    elif "基础连接已经关闭" in err:
        # 说明目标服务使用了https服务，但是SSL证书不安全，但隧道转发是正常的
        return True
    else:
        return True


def ssh_远程转发(remote_ip, remote_port, local_port, login_user):
    """
    该函数需要在提供服务的计算机上执行，一般提供服务的计算机为内网计算机，没有公网ip，该方法将内网计算机上的服务映射到公网ip服务器的端口上。
    Args:
        remote_ip: 远程服务器的ip，一般为租赁的云服务器，具有公网ip
        remote_port: 远程服务器开放的端口
        local_port: 内网计算机提供服务的端口
        login_user: 远程服务器的ssh登录用户名

    Returns:

    """

    cmd = f"ssh -R 0.0.0.0:{remote_port}:127.0.0.1:{local_port} {login_user}@{remote_ip} -o ServerAliveInterval=60 -o ServerAliveCountMax=3"
    if not check_ssh_tunnel_exist(remote_ip, remote_port):  # 如果端口转发不存在
        runCMD(cmd)


if __name__ == "__main__":
    # ssh_远程转发("182.43.65.44", "5000", "20010", "root")
    print(check_ssh_tunnel_exist("101.37.118.81", "20011"))



import torch

dtype = torch.float
device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
print(device)

a = torch.randn(2, 3, device=device, dtype=dtype)  # 创建一个 2x3 的随机张量
b = torch.randn(2, 3, device=device, dtype=dtype)  # 创建另一个 2x3 的随机张量
print("张量 a:")
print(a)

print("张量 b:")
print(b)

# 逐元素相乘并输出结果
print("a 和 b 的逐元素乘积:")
print(a * b)

# 输出张量 a 所有元素的总和
print("张量 a 所有元素的总和:")
print(a.sum())


"""
本模块用于更新mysql服务器中的数据
"""
import datetime
import pandas as pd
from yangke.common.config import logger
from yangke.base import execute_function_every_day
from yangke.stock.dataset.mairuiData import StockData
from yangke.stock.dataset.storage import Storage
import yangke.stock.dataset.baostockData as bsd


class UpdateDataBase:
    def __init__(self, kind='dataset', ip=None, port=None, user=None, passwd=None, db=None):
        """
        更新股票数据至数据库
        :param kind: 可取值mysql，表示存储到mysql数据库，sqlite表示sqlite数据库，其他字符串表示存储到运行路径下的文件夹
        """
        self.storage = Storage(kind, user, passwd, ip, port, db)
        # 新增：检查存储初始化是否成功
        if self.storage.engine is None:
            logger.error(f"存储初始化失败：{self.storage.error}")
            # 可根据需求选择退出程序或执行其他降级操作
            raise RuntimeError("存储初始化失败，请检查数据库权限或配置")
        self.ds: StockData = StockData()
        # 一次性获取所有股票当天的数据，一天多次获取可能返回None，是服务器限制了请求次数
        res = self.ds.get_daily_all(need_json=True)
        if res is not None:
            d1, _ = res
            self.storage.update_daily_all(d1)

    def update_all_holidays(self):
        """
        更新服务器上的假日信息数据，手动运行，一次更新10年的假日信息最好
        """
        holiday = bsd.get_holiday(start_date=datetime.date(
            1990, 12, 19), end_date=datetime.date(2000, 1, 1))
        self.storage.update_holiday_table(holiday)

    def update(self):
        """
        更新一次股票数据
        每次执行时，会自动检测当天所属的年份的假日数据是否存在，不存在增更新当年的假日数据

        """
        # ----------------------------- 首先更新假期表 -----------------------------------------
        # 只有tushare和baostock有假期数据
        if self.storage.need_update_holiday_table():
            holiday = bsd.get_holiday()  # 获取今年的holiday数据
            self.storage.update_holiday_table(holiday)

        # ----------------------------- 更新全部股票列表数据 ------------------------------------
        all_stocks: pd.DataFrame | None = None
        if self.storage.need_update_all_stocks_table():  # 从网络获取
            all_stocks = self.ds.get_all_stock_basic_info()
            self.storage.update_all_stocks_table(all_stocks)
        if all_stocks is None:  # 获取失败，从自建数据库获取
            all_stocks = self.storage.get_all_stock_basic_info()  # 从数据库存储中读取所有的股票信息

        last_day_desired = self.storage.get_previous_working_day(
            include_time=True)
        pre_day = self.storage.get_working_day_before_day(last_day_desired)
        if (isinstance(self.storage.daily_all_data, list)
                and len(self.storage.daily_all_data) > 1
                and pd.to_datetime(self.storage.daily_all_data[0].get('t')).date() == last_day_desired):
            # 如果daily_all_data已经是当天的数据，则不在更新
            daily_all = pd.DataFrame(self.storage.daily_all_data)
            daily_all.rename(columns={
                "t": "trade_date",
                "o": "open", "h": "high",
                "l": "low", "p": "close",
                "v": "vol", "cje": "amount",
                "hs": "换手率",
                "sz": "市值",
                "lt": "流通市值",
                "pe": "市盈率",
                "sjl": "市净率",
                "dm": "symbol"
            }, inplace=True)
        else:
            d1, daily_all = self.ds.get_daily_all(
                need_json=True)  # 一次性获取所有股票当天的数据
            self.storage.update_daily_all(d1)

        for index, row in all_stocks.iterrows():
            symbol = row['symbol']
            logger.info(f"开始更新{symbol}的日线数据，进度{index}/{all_stocks.shape[0]}")
            try:
                self.storage.create_stock_table_day(symbol)
                if self.storage.need_update_daily(symbol):  # 首先判断是否需要更新数据
                    # 其次判断表中最后一条数据的日期，如果只是缺少一条数据，只需要获取最后一天的数据即可
                    last_date = self.storage.get_last_date_of_daily_table(
                        symbol)
                    if last_date == pre_day:
                        d = daily_all[daily_all['symbol'] == symbol]
                        if len(d) == 1 and pd.to_datetime(d["trade_date"]).iloc[0].date() == last_day_desired:
                            daily_data = d
                        else:
                            daily_data = self.ds.get_single_day_data(symbol)
                    else:
                        daily_data = self.ds.get_daily(symbol, end_date=last_day_desired)
                    if daily_data is not None:
                        self.storage.update_daily(symbol, daily_data)
            except TimeoutError:
                logger.debug(f"获取{symbol}的日线数据失败，跳过")
        logger.debug("数据更新完成!")

    def start(self, daemon=True, test_service=False):
        if not test_service:
            execute_function_every_day(
                self.update, hour=16, minute=0, daemon=False)
        self.storage.start_rest_service(daemon=daemon)


if __name__ == "__main__":
    udb = UpdateDataBase(kind='mysql')
    udb.update()
    udb.start(test_service=True)



from yangke.ebsilon.ebsilon import EbsApp, EbsModel, EbsProfile
from yangke.web.flaskserver import start_server_app
from yangke.common.config import logger
from typing import Optional
import pythoncom

ebs = EbsApp()


def EbsilonAvailable(args):
    _ = ebs.get_application_string()
    return _


def GenerateTCurve(args):
    model_path = args.get("modelPath")
    model = ebs.open(model_path)
    profile = model.get_profile("母体工况")
    root_profile = model.get_profile("设计")
    success, coped_profile = profile.copy(profile, False, root_profile)
    coped_profile.change_name("大气温度")
    children = [coped_profile]
    for i in range(1, 9):
        children.append(coped_profile.new_child())
        children[i].change_name(f"大气温度{i}")

    model.save_as(r"E:\new.ebs", True)


def solve(args):
    action = args.get('Action')  # 因为下方use_action=True，所以这里的action必然有值，避免eval函数出错
    result = eval("{}(args)".format(action))
    return result


if __name__ == "__main__":
    start_server_app(deal=solve, port=20011, host="0.0.0.0", single_thread=True)  # 必须以单线程模式启动，否则EbsApp难以运行



# 用于测量屏幕上鼠标点击位置的距离

import pyWinhook  # pip install pywinhook
from pyWinhook.HookManager import MouseEvent, KeyboardEvent
import pythoncom

from yangke.common.KeysTransfer import KeysTransfer
from yangke.common.config import logger


# collapse
def shield_hot_key():
    import PyHook3  # 先安装swig.exe，配置环境变量后，再安装pip install PyHook3

    hm = PyHook3.HookManager()

    # 鼠标事件处理函数
    def OnMouseEvent(event):
        print('MessageName:', event.MessageName)  # 事件名称
        print('Message:', event.Message)  # windows消息常量
        print('Time:', event.Time)  # 事件发生的时间戳
        print('Window:', event.Window)  # 窗口句柄
        print('WindowName:', event.WindowName)  # 窗口标题
        print('Position:', event.Position)  # 事件发生时相对于整个屏幕的坐标
        print('Wheel:', event.Wheel)  # 鼠标滚轮
        print('Injected:', event.Injected)  # 判断这个事件是否由程序方式生成，而不是正常的人为触发。
        print('---')

        # 返回True代表将事件继续传给其他句柄，为False则停止传递，即被拦截
        return True

    # 键盘事件处理函数
    def OnKeyboardEvent(event):
        print('MessageName:', event.MessageName)  # 同上，共同属性不再赘述
        print('Message:', event.Message)
        print('Time:', event.Time)
        print('Window:', event.Window)
        print('WindowName:', event.WindowName)
        print('Ascii:', event.Ascii, chr(event.Ascii))  # 按键的ASCII码
        print('Key:', event.Key)  # 按键的名称
        print('KeyID:', event.KeyID)  # 按键的虚拟键值
        print('ScanCode:', event.ScanCode)  # 按键扫描码
        print('Extended:', event.Extended)  # 判断是否为增强键盘的扩展键
        print('Injected:', event.Injected)
        print('Alt', event.Alt)  # 是某同时按下Alt
        print('Transition', event.Transition)  # 判断转换状态
        print('---')

        # 同上
        return False

    # 绑定事件处理函数
    # hm.MouseAllButtonsDown = OnMouseEvent  # 将OnMouseEvent函数绑定到MouseAllButtonsDown事件上
    hm.KeyDown = OnKeyboardEvent  # 将OnKeyboardEvent函数绑定到KeyDown事件上
    hm.KeyUp = OnKeyboardEvent

    # hm.HookMouse()  # 设置鼠标钩子
    hm.HookKeyboard()  # 设置键盘钩子

    # 循环获取消息
    pythoncom.PumpMessages()


class MouseKeyManager:
    keyIsPressed = False

    def __init__(self):
        """
        全局鼠标键盘响应事件
        """
        self.hm = pyWinhook.HookManager()
        self.on_key_event = None
        self.start_pos = None
        self.end_pos = None
        self.pressed_key = None
        self.dragging = False
        self.ignore_keys = set()

    def onKeyDown(self, event):
        print(f"按下键盘：{event.Key}")
        if self.keyIsPressed:
            return True
        self.keyIsPressed = True
        return True

    def onKeyUp(self, event):
        self.keyIsPressed = False
        print(str(event.Key) + ' is released')
        return True

    def mouseup(self, event: MouseEvent):
        self.end_pos = event.Position
        dx, dy = self.end_pos[0] - self.start_pos[0], self.end_pos[1] - self.start_pos[1]
        if dx > 1 or dy > 1:
            print(f"按下鼠标：{self.start_pos}")
            print(f"弹起鼠标：{event.Position}")
            # if self.dragging:
            print(f"dx, dy = {(self.end_pos[0] - self.start_pos[0], self.end_pos[1] - self.start_pos[1])}")
            self.dragging = False
        return True

    def mousedown(self, event: MouseEvent):
        self.start_pos = event.Position
        return True

    def _ignore_key(self, key_event: KeyboardEvent):
        """
        拦截系统按键，拦截self.ignore_keys中的按键，并调用self.on_key_event回调函数，如果没有回调函数，则只是屏蔽相应按键
        """
        if key_event.Key and key_event.Key.lower() in self.ignore_keys:
            if self.on_key_event:
                return self.on_key_event(key_event)
            else:
                return False
        else:
            return True

    def shield_hot_keys(self, keys, on_key_event=None):
        """
        屏蔽系统热键，当系统热键触发时，调用on_key_event方法
        键名参考KeysTransfer中的default键名
        只能调用一次
        """
        if isinstance(keys, str):
            key = KeysTransfer(source="default", destination="pywinhook_name").transfer(keys)
            self.ignore_keys.append(key.lower())
        elif isinstance(keys, list):
            for key in keys:
                key = KeysTransfer(source="default", destination="pywinhook_name").transfer(key) or ""
                self.ignore_keys.add(key.lower())

        self.on_key_event = on_key_event
        self.hm.KeyUp = self._ignore_key
        self.hm.KeyDown = self._ignore_key
        self.hm.HookKeyboard()
        self.daemon()

    def stop_shield_keyboard(self):
        self.hm.UnhookKeyboard()

    def stop_shield_mouse(self):
        self.hm.UnhookMouse()

    @staticmethod
    def daemon():
        pythoncom.PumpMessages()


if __name__ == '__main__':
    mkm = MouseKeyManager()
    mkm.shield_hot_keys(["win", "PrintScreen"])
    mkm.daemon()
